@@ -6,6 +6,9 @@
 #include <sisocks.h>
 #include <string.h>
 #include <stdio.h>
+#include <sys/stat.h>
+#include "bsdcmpt.h"
+#include <time.h>
 
 /* size of the line buffer for each worker (request and header only)
  * requests that have longer headers will be rejected with 413
@@ -61,7 +64,7 @@ struct args {
 	/* the following entries are not populated by Rserve but can be used by server implemetations */
 	char *buf, *sbuf;
 	int   ver, bp, bl, sp, sl, flags;
-	long  l1, l2;
+	size_t l1, l2;
 	/* The following fields are informational, populated by Rserve */
     SAIN sa;
     int ucix;
@@ -128,6 +131,29 @@ static SEXP collect_buffers(struct buffer *buf) {
     return res;
 }
 
+/* convert doubly-linked buffers into one big string,
+   caller must free, may return NULL */
+static char *collect_buffers_c(struct buffer *buf) {
+    char *res, *dst;
+    int len = 0;
+    if (!buf) return 0;
+    while (buf->prev) { /* count the total length and find the root */
+		len += buf->length;
+		buf = buf->prev;
+    }
+    res = malloc(len + buf->length + 1);
+	if (!res)
+		return res;
+    dst = res;
+    while (buf) {
+		memcpy(dst, buf->data, buf->length);
+		dst += buf->length;
+		buf = buf->next;
+    }
+	*dst = 0; /* terminate */
+    return res;
+}
+
 static void free_args(args_t *c)
 {
     DBG(printf("finalizing worker %p\n", (void*) c));
@@ -312,6 +338,254 @@ static void fin_request(args_t *c) {
 		c->attr |= CONNECTION_CLOSE;
 }
 
+static SEXP s_http_request_fn;
+
+static void http_set_request_fn(SEXP sFn) {
+	s_http_request_fn = sFn ? sFn : install(".http.request");
+}
+
+/* API */
+SEXP Rserve_set_http_request_fn(SEXP sFn) {
+	http_set_request_fn((sFn == R_NilValue) ? 0 : sFn);
+	return s_http_request_fn;
+}
+
+typedef struct http_static {
+	struct http_static *next;
+	char *prefix, *path, *index;
+	int prefix_len;
+	int flags;
+} http_static;
+
+static http_static *http_statics;
+
+void *http_add_static_handler(const char *prefix, const char* path, const char *index, int flags) {
+	http_static *h, *hl;
+	h = (http_static*) malloc(sizeof(http_static));
+    if (!h)
+        return 0;
+    h->next = 0;
+    h->prefix = strdup(prefix ? prefix : "");
+    h->path = strdup(path ? path : "");
+    h->index = index ? strdup(index) : 0;
+    h->prefix_len = strlen(h->prefix);
+    h->flags = flags;
+    if (!http_statics) {
+        http_statics = h;
+    } else {
+        hl = http_statics;
+        while (hl->next)
+            hl = hl->next;
+        hl->next = h;
+    }
+	return (void*) h;
+}
+
+static void http_free_static_handler(void *hs) {
+	http_static *h = (http_static*) hs;
+	if (h->prefix)
+		free(h->prefix);
+	if (h->path)
+		free(h->path);
+	if (h->index)
+		free(h->index);
+	free(h);
+}
+
+void http_rm_static_handler(void *hs) {
+	http_static *h = (http_static*) hs;
+	if (!h)
+		return;
+	if (http_statics == h)
+		http_statics = h->next;
+	else {
+		http_static *h = http_statics;
+		while (h && h->next != hs)
+			h = h->next;
+		if (h)
+			h->next = h->next->next;
+		else
+			return; /* not found */
+	}
+	http_free_static_handler(hs);
+}
+
+void http_rm_all_static_handlers(void) {
+	while (http_statics) {
+		http_static *h = http_statics;
+		http_statics = h->next;
+		http_rm_static_handler(h);
+	}
+}
+
+/** R API **/
+
+SEXP Rserve_http_add_static(SEXP sPrefix, SEXP sPath, SEXP sIndex, SEXP sFlags) {
+	int n = 0;
+	http_static *h;
+	if (TYPEOF(sPrefix) != STRSXP || LENGTH(sPrefix) != 1)
+		Rf_error("Invalid prefix, must be a string");
+	if (TYPEOF(sPath) != STRSXP || LENGTH(sPath) != 1)
+		Rf_error("Invalid path, must be a string");
+	if (!((TYPEOF(sIndex) == STRSXP && LENGTH(sPath) == 1) || sIndex == R_NilValue))
+		Rf_error("Invalid index, must be NULL or a string");
+	
+	if (!http_add_static_handler(strdup(CHAR(STRING_ELT(sPrefix, 0))),
+								 strdup(CHAR(STRING_ELT(sPath, 0))),
+								 (sIndex == R_NilValue) ? 0 : strdup(CHAR(STRING_ELT(sIndex, 0))),
+								 Rf_asInteger(sFlags)))
+		Rf_error("Cannot allocate handler structure.");
+	h = http_statics;
+	while (h) {
+		n++;
+		h = h->next;
+	}
+	return Rf_ScalarInteger(n);
+}
+
+SEXP Rserve_http_rm_all_statics(void) {
+	http_rm_all_static_handlers();
+	return Rf_ScalarLogical(1);
+}
+
+static char http_tmp[512];
+
+/* 0 = ok, sent
+  -1 = failure during transmission, have to close
+  -2 = file cannot be read (nothing sent) */
+static int http_send_file(args_t *c, const char *fn, size_t fsz, int size_valid, const char *preamble) {
+	char *fbuf;
+	char buf[64];
+	FILE *f = fopen(fn, "rb");
+	if (!f)
+		return -2;
+	/* FIXME: on Windows this is limited to 2Gb */
+	if (!size_valid) {
+		fseek(f, 0, SEEK_END);
+		fsz = (size_t) ftell(f);
+		fseek(f, 0, SEEK_SET);
+	}
+	if (preamble)
+		send_http_response(c, preamble);
+	snprintf(buf, sizeof(buf), "\r\nContent-length: %llu\r\n\r\n", (unsigned long long) fsz);
+	send_response(c, buf, strlen(buf));
+#define HTTP_SEND_FILE_CHUNK (1024*1024)
+	if (fsz && c->method != METHOD_HEAD) {
+		fbuf = (char*) malloc(HTTP_SEND_FILE_CHUNK);
+		if (fbuf) {
+			while (fsz > 0 && !feof(f)) {
+				int rd = (fsz > HTTP_SEND_FILE_CHUNK) ? HTTP_SEND_FILE_CHUNK : fsz;
+				if (fread(fbuf, 1, rd, f) != rd) {
+					free(fbuf);
+					fclose(f);
+					return -1;
+				}
+				send_response(c, fbuf, rd);
+				fsz -= rd;
+			}
+			free(fbuf);
+		} else { /* allocation error - get out */
+			fclose(f);
+			return -1;
+		}
+	}
+	fclose(f);
+	return 0;
+}
+
+/* returns a pointer to the beginning of a value for a given header
+   field or NULL if not present. */
+static const char *get_header(const char *headers, const char *name) {
+    const char *c = headers, *e;
+    int name_len = strlen(name);
+    if (!c) return 0;
+    while (*c && (e = strchr(c, '\n'))) {
+	const char *v = strchr(c, ':');
+	if (v && (v < e) && (v - c == name_len)) {
+	    int i;
+	    for (i = 0; i < name_len; i++)
+		if ((name[i] & 0xdf) != (c[i] & 0xdf))
+		    break;
+	    if (i == name_len) {
+		v++;
+		while (*v == '\t' || *v == ' ')
+		    v++;
+		return v;
+	    }
+	}
+	while (*e == '\n' || *e == '\t') e++;
+	c = e;
+    }
+    return 0;
+}
+
+static const char *infer_content_type(const char *fn) {
+    const char *x = fn ? strrchr(fn, '.') : 0;
+	char ext[8]; /* lower-case version of the extension */
+	int i = 0;
+    if (!x) return "application/octet-stream";
+	x++;
+	while (i < 7 && *x) {
+		ext[i++] = (*x >= 'A' && *x <= 'Z') ? (*x | 0x20) : *x;
+		x++;
+	}
+	ext[i] = 0;
+
+    if (!strcmp(ext, "svg")) return "image/svg+xml";
+    if (!strcmp(ext, "js"))  return "application/javascript";
+    if (!strcmp(ext, "css")) return "text/css";
+    if (!strcmp(ext, "html")) return "text/html";
+    if (!strcmp(ext, "txt")) return "text/plain";
+    if (!strcmp(ext, "png")) return "image/png";
+    if (!strcmp(ext, "jpeg") || !strcmp(ext, "jpg")) return "image/jpeg";
+    if (!strcmp(ext, "md")) return "text/markdown";
+	if (!strcmp(ext, "json")) return "application/json";
+    return "application/octet-stream";
+}
+
+/* removes .. and . segments, control characters.
+   non-ASCII characters are retained */
+static void sanitize_path(char *path) {
+	unsigned char *src = (unsigned char*) path, *dst = (unsigned char*) path;
+	int pos = 0;
+	DBG(fprintf(stderr, "path: '%s' sanitizing\n", path));
+	while (*src) {
+		if (!pos) {
+			if (src[0] == '.') {
+				/* .. */
+				if (src[1] == '.' && (!src[2] || src[2] == '/')) {
+					src += 2;
+					continue;
+				}
+				/* . (harmless, but we still remove it) */
+				if (!src[1] || src[1] == '/') {
+					src++;
+					continue;
+				}
+			}
+		}
+		if (*src == '/') {
+			if (pos) { /* not a repeated / */
+				*(dst++) = *(src++);
+				pos = 0;
+			} else /* repeated, just ignore */
+				src++;
+			continue;
+		}
+		pos++;
+		if (*src >= 32)
+			*(dst++) = *(src++);
+		else
+			src++; /* skip (we shold probably fail, though) */
+	}
+	*dst = 0;
+	DBG(fprintf(stderr, "      '%s'\n", path));
+}
+
+/* from date.c */
+char *posix2http(double ts); /* Note: returned buffer is static */
+double http2posix(const char *c);
+
 /* process a request by calling the httpd() function in R */
 static void process_request(args_t *c)
 {
@@ -333,7 +607,97 @@ static void process_request(args_t *c)
 		*(s++) = 0;
 		query = s;
     }
+	if (!s_http_request_fn)
+		s_http_request_fn = install(".http.request");
     uri_decode(c->url); /* decode the path part */
+
+	/* it would be nice if we could re-use the code from forward.c,
+	   but for now they are separate. FIXME: move the forward
+	   static handling into an http handler, make the R API call
+	   another handler and then replace the code here with the
+	   handler code from forward.c */
+	if (http_statics) {
+		http_static *hs = http_statics;
+		while (hs) {
+			DBG(fprintf(stderr, "http_static: match '%s' and '%s'\n", c->url, hs->prefix));
+			if (!strncmp(hs->prefix, c->url, hs->prefix_len)) {
+				struct stat st;
+				const char *fn = c->url + hs->prefix_len;
+				size_t path_len = strlen(hs->path);
+				int found = 0;
+				/* 7 padding is plenty - we need up to two '/' and one NUL */
+				if (strlen(fn) + path_len +
+					(hs->index ? strlen(hs->index) : 0) +
+					7 > sizeof(http_tmp)) {
+					send_http_response(c, " 414 Path too long\r\nContent-type: text/plain\r\nContent-length: 14\r\n\r\nPath too long\n");
+					fin_request(c);
+					return;
+				}
+				strcpy(http_tmp, hs->path);
+				/* make sure there is always / between the path and the filename
+				   (except if path="" for backward compatibility meaning ".") */
+				if (path_len > 0 &&
+					(http_tmp[path_len - 1] != '/' || *fn != '/'))
+					http_tmp[path_len++] = '/';
+				strcpy(http_tmp + path_len, fn);
+				/* only sanitize the part coming from the request */
+				sanitize_path(http_tmp + path_len);
+				if (!stat(http_tmp, &st)) { /* path exists */
+					if (st.st_mode & S_IFDIR) { /* if it is a directory, we only accept it if index exists */
+						if (hs->index) {
+							size_t http_tmp_len = strlen(http_tmp);
+							if (http_tmp_len > 0 && http_tmp[http_tmp_len - 1] != '/')
+								http_tmp[http_tmp_len++] = '/';
+							strcpy(http_tmp + http_tmp_len, hs->index);
+							DBG(fprintf(stderr, " - target is a directory, try '%s'\n", http_tmp));
+							if (!stat(http_tmp, &st))
+								found = 1;
+						}
+					} else if (st.st_mode & S_IFREG) /* otherwise we only accept regular files */
+						found = 1;
+				}
+				if (!found) {
+					DBG(fprintf(stderr, " - '%s' not found\n", http_tmp));
+					if (hs->flags & HSF_STOP) {
+						send_http_response(c, " 404 Not found\r\nContent-type: text/plain\r\nContent-length: 10\r\n\r\nNot found\n");
+						fin_request(c);
+						return;
+					}
+				} else { /* file found */
+					int not_modified = 0;
+					/* check for conditional GET */
+					if (c->headers) {
+						char *h = collect_buffers_c(c->headers);
+						const char *if_mod = get_header(h, "if-modified-since");
+						if (if_mod) {
+							double since = http2posix(if_mod);
+							if (since >= MTIME(st))
+								not_modified = 1;
+						}
+						free(h);
+					}
+					if (not_modified) {
+						send_http_response(c, " 304 Not modified\r\nCache-Control: no-cache\r\n\r\n");
+					} else {
+						char buf[196];
+						double ts = (double) time(0);
+						snprintf(buf, sizeof(buf), " 200 OK\r\nCache-Control: no-cache\r\nContent-type: %s\r\nLast-Modified: %s",
+								 infer_content_type(http_tmp), posix2http((MTIME(st) > ts) ? ts : MTIME(st)));
+						int res = http_send_file(c, http_tmp, (size_t) st.st_size, 1, buf);
+						if (res == -2) { /* cannot open */
+							send_http_response(c, " 403 Forbidden\r\nContent-Type: text/plain\r\nContent-length: 10\r\n\r\nForbidden\n");
+						}
+						if (res == -1) /* transfer error */
+							c->attr |= CONNECTION_CLOSE;
+					}
+					/* all paths here sent the response */
+					fin_request(c);
+					return;
+				}
+			}
+			hs = hs->next;
+		}
+	}
     {   /* construct "try(httpd(url, query, body, headers), silent=TRUE)" */
 		SEXP sTrue = PROTECT(ScalarLogical(TRUE));
 		SEXP sBody = PROTECT(parse_request_body(c));
@@ -342,7 +706,7 @@ static void process_request(args_t *c)
 		SEXP sArgs = PROTECT(list4(mkString(c->url), sQuery, sBody, sReqHeaders));
 		SEXP sTry = install("try");
 		SEXP y, x = PROTECT(lang3(sTry,
-								  LCONS(install(".http.request"), sArgs),
+								  LCONS(s_http_request_fn, sArgs),
 								  sTrue));
 		SET_TAG(CDR(CDR(x)), install("silent"));
 		DBG(Rprintf("eval(try(.http.request('%s'),silent=TRUE))\n", c->url));
@@ -404,7 +768,7 @@ static void process_request(args_t *c)
 				if (code == 200)
 					send_http_response(c, " 200 OK\r\nContent-type: ");
 				else {
-					sprintf(buf, "%s %d Code %d\r\nContent-type: ", HTTP_SIG(c), code, code);
+					snprintf(buf, sizeof(buf)-1, "%s %d Code %d\r\nContent-type: ", HTTP_SIG(c), code, code);
 					send_response(c, buf, strlen(buf));
 				}
 				send_response(c, ct, strlen(ct));
@@ -424,52 +788,22 @@ static void process_request(args_t *c)
 					(!strcmp(CHAR(STRING_ELT(xNames, 0)), "file") || (is_tmp = !strcmp(CHAR(STRING_ELT(xNames, 0)), "tmpfile"))))
 					fn = cs;
 				if (fn) {
-					char *fbuf;
-					FILE *f = fopen(fn, "rb");
-					long fsz = 0;
-					if (!f) {
+					int res = http_send_file(c, fn, 0, 0, 0);
+					if (res == -2) { /* cannot open */
 						send_response(c, "\r\nContent-length: 0\r\n\r\n", 23);
 						UNPROTECT(7);
 						fin_request(c);
 						return;
+					} else if (res) {
+						/* force close since the transfer failed */
+						c->attr |= CONNECTION_CLOSE;
 					}
-					fseek(f, 0, SEEK_END);
-					fsz = ftell(f);
-					fseek(f, 0, SEEK_SET);
-					sprintf(buf, "\r\nContent-length: %ld\r\n\r\n", fsz);
-					send_response(c, buf, strlen(buf));
-					if (c->method != METHOD_HEAD) {
-						fbuf = (char*) malloc(32768);
-						if (fbuf) {
-							while (fsz > 0 && !feof(f)) {
-								int rd = (fsz > 32768) ? 32768 : fsz;
-								if (fread(fbuf, 1, rd, f) != rd) {
-									free(fbuf);
-									UNPROTECT(7);
-									c->attr |= CONNECTION_CLOSE;
-									fclose(f);
-									if (is_tmp) unlink(fn);
-									return;
-								}
-								send_response(c, fbuf, rd);
-								fsz -= rd;
-							}
-							free(fbuf);
-						} else { /* allocation error - get out */
-							UNPROTECT(7);
-							c->attr |= CONNECTION_CLOSE;
-							fclose(f);
-							if (is_tmp) unlink(fn);
-							return;
-						}
-					}
-					fclose(f);
 					if (is_tmp) unlink(fn);
 					UNPROTECT(7);
 					fin_request(c);
 					return;
 				}
-				sprintf(buf, "\r\nContent-length: %u\r\n\r\n", (unsigned int) strlen(cs));
+				snprintf(buf, sizeof(buf)-1, "\r\nContent-length: %u\r\n\r\n", (unsigned int) strlen(cs));
 				send_response(c, buf, strlen(buf));
 				if (c->method != METHOD_HEAD)
 					send_response(c, cs, strlen(cs));
@@ -483,7 +817,7 @@ static void process_request(args_t *c)
 				if (code == 200)
 					send_http_response(c, " 200 OK\r\nContent-type: ");
 				else {
-					sprintf(buf, "%s %d Code %d\r\nContent-type: ", HTTP_SIG(c), code, code);
+					snprintf(buf, sizeof(buf)-1, "%s %d Code %d\r\nContent-type: ", HTTP_SIG(c), code, code);
 					send_response(c, buf, strlen(buf));
 				}
 				send_response(c, ct, strlen(ct));
@@ -495,7 +829,7 @@ static void process_request(args_t *c)
 						send_response(c, hs, strlen(hs));
 					}
 				}
-				sprintf(buf, "\r\nContent-length: %u\r\n\r\n", LENGTH(y));
+				snprintf(buf, sizeof(buf)-1, "\r\nContent-length: %u\r\n\r\n", LENGTH(y));
 				send_response(c, buf, strlen(buf));
 				if (c->method != METHOD_HEAD)
 					send_response(c, (char*) cs, LENGTH(y));
